// ================================================================
// GenomeMixer    (c) 2003, Alex Williams
// All rights reserved.
// Send email correspondence to Robert W. Williams <rwilliam@nb.utmem.edu>
// at the University of Tennesse, Memphis

// This project was funded jointly by the National Institute of Mental Health, National Institute on Drug Abuse,
// and the National Science Foundation (award P20-MH 62009 to KFM and RW)

// This source code may be used according to the terms of the GNU General Public License (GPL).
// See www.gnu.org for specifics. Essentially, the source may be freely used as long as the source code
// for whatever program the source is also made freely available. If you want to use this source
// in a closed-source program, you should email us for licensing options.

// Disclaimer: This code comes with no warranty at all, and there is no guarantee that it will
// work as claimed in the documentation. Hopefully, it will be useful anyway.

// Note that GenomeMixer makes use of the Trolltech Qt toolkit for its window-managing needs.
// There is very little use of Qt data structures (aside from QString) in the actual processing code,
// but **you will need to have Qt installed in order to compile this code.** Check <www.trolltech.com>
// for details; there should be a free GPL-software-only version of Qt available on the Trolltech
// web site.

// Any redistributions of this software in code or binary form must include the
// above copyright notice, licensing information, disclaimer, and this notice itself.
// Basically, all the comments at the top of this file must be retained.
// ================================================================


/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
**
** If you wish to add, delete or rename functions or slots use
** Qt Designer which will update this file, preserving your code. Create an
** init() function in place of a constructor, and a destroy() function in
** place of a destructor.
*****************************************************************************/

// TABS
const int tab_chr = 0;
const int tab_markers = 1;
const int tab_purebreds = 2;
const int tab_scheme = 3;

// CHROMOSOME
const int chr_MB_col = 0;
const int chr_cM_col = 1;

// MARKERS
const int markers_chr_col	= 0;
const int markers_name_col	= 1;
const int markers_bases_col	= 2;
const int markers_cM_col	= 3;

// PUREBREDS
const int purebreds_name_col		= 0;
const int purebreds_symbol_col	= 1;

// SCHEME
const int pulldown_purebred = 0;
const int pulldown_from_cell = 1;


void SettingsDialog::init() {
	
	mainMixPtr = NULL;
	
	settingsChangedAtAll = false; // <-- gets changed when someone hits "apply"
	markersChangedAtAll = false; // <-- gets changed when someone hits "apply"    
	settingsDirty = true; // <-- true, since we just want to make sure things get loaded
	markersDirty = true;
	
	// Everything else important is done in setUpSettings, which connects a GenomeMixer object
	
	// CHROMOSOME
	tableChr->setColumnStretchable(chr_MB_col, true);
	tableChr->setColumnStretchable(chr_cM_col, true);
	
	tableSexChr->setColumnStretchable(chr_MB_col, true);
	tableSexChr->setColumnStretchable(chr_cM_col, true);
	tableSexChr->setRowStretchable(0, true); // <-- X chr row
	tableSexChr->setRowStretchable(1, true); // <-- Y chr row
	
	// MARKERS
	tableMarkers->setColumnWidth(markers_chr_col, 30);
	tableMarkers->setColumnWidth(markers_cM_col, 60);
	tableMarkers->setColumnStretchable(markers_chr_col	, true);
	tableMarkers->setColumnStretchable(markers_name_col	, true);
	tableMarkers->setColumnStretchable(markers_bases_col	, true);
	tableMarkers->setColumnStretchable(markers_cM_col	, true);
	
	// PUREBREDS
	tablePurebreds->setColumnStretchable(purebreds_name_col, true);
	tablePurebreds->setColumnStretchable(purebreds_symbol_col, true);
	
	textDisplayInfo->setText(" \n "); // clear the "display information" text field
	
	// Set the maximum size for the array (#generations and #cages)
	spinSchemeCages->setMaxValue(kTABLE_SIZE);
	spinSchemeGenerations->setMaxValue(kTABLE_SIZE);
}


void SettingsDialog::setUpSettings(GenomeMixer *theMainPtr) {    
	
	assert(theMainPtr != NULL);
	
	mainMixPtr = theMainPtr;
	
	for (int i = 0; i < kTABLE_SIZE; i++) {
		for (int k = 0; k < kTABLE_SIZE; k++) {
			tableEntries[i][k] = NULL; // start the tableEntries off as NULL
		}
	}    
	
	textDisplayInfo->setText(" \n "); // clear the "display information" text field   
	
	settingsChangedAtAll = false; // <-- gets changed when someone hits "apply"
	markersChangedAtAll = false; // <-- gets changed when someone hits "apply"
	settingsDirty = true;
	markersDirty = true;
	reloadEverything();
}



void SettingsDialog::reloadEverything() {
	if (settingsDirty) {
		// CHROMOSOME INITIALIZATION
		reloadChrData(); // just loads in the chromosome data
		updateChromosomeInfo(); // updates the info about how many chromosomes we have, etc
		// PUREBRED INITIALIZATION
		reloadPurebredData();
		updatePurebredInfo();
		// SCHEME INITIALIZATION
		reloadSchemeData();
		updateSchemeInfo();
	}
	
	if (markersDirty) {
		// MARKER INITIALIZATION
		reloadMarkerData();
		updateMarkerInfo();
	}
	
	// EVERYTHING ELSE
	makeAllClean(); // now that we have reloaded everything, it is "clean"
}




void SettingsDialog::doneButtonWasPressed() {
	applyChanges();
	accept();
}



void SettingsDialog::theCurrentTabChanged() {
	// The way it works now is that switching between tabs applies the changes, too
	if (settingsDirty || markersDirty) {
		applyChanges();
		textDisplayInfo->setText("The previous tab's settings were applied when you switched tabs.\n(They were not saved to disk, however.)");
	}
}



void SettingsDialog::renumberTable(QTable *table, int start, int finishNonInclusive) {
	for (int i = start; i < finishNonInclusive; i++) {
		// Set the labels to 1, 2, 3, ... 
		table->verticalHeader()->setLabel(i, QIconSet(NULL), QString::number(i+1)); // Re-number
	}
}

void SettingsDialog::makeAllClean() {
	settingsDirty = false; // now it's clean
	markersDirty = false; // now it's clean
	updateSaveDiscardButtons();
}

void SettingsDialog::makeCurrentPaneDirty() {
	// Depending on which pane we are in, we set either the markers or the settings to "dirty"
	
	switch (settingsTabWidget->currentPageIndex()) {
		case tab_chr:  // fall through to next "case"...
		case tab_purebreds:  // fall through to next "case"...
	case tab_scheme:
		textDisplayInfo->setText("Settings have been changed.\n");
		settingsDirty = true;
		break;
		
	case tab_markers:
		textDisplayInfo->setText("Markers have been changed.\n");
		markersDirty = true;
		break;
		
	default:
		cerr << "It looks like we are inside a tab that we didn't know about. Very suspicious." << endl;
		break;
	}
	
	updateSaveDiscardButtons();
}


void SettingsDialog::updateSaveDiscardButtons() {
	// If things are "dirty," then the buttons will be enabled, so you can save or revert.
	// If things are clean (unchanged), then the buttons will be disabled
	//buttonApplyChanges->setEnabled(settingsDirty || markersDirty);
	buttonDiscardChanges->setEnabled(settingsDirty || markersDirty);
	
}



void SettingsDialog::applyChanges() {
	// Apply the changes (save over the current settings in RAM, but don't save to disk).
	if (settingsDirty) {
		settingsChangedAtAll = true; // we changed the settings, so now the settings file should be save-able	
		chrTableApply();
		purebredTableApply();
		schemeTableApply();
	}
	
	if (markersDirty) {
		markersChangedAtAll = true; // we changed them, so now the marker file should be save-able		
		markersTableApply();
	}
	
	if (settingsDirty || markersDirty) {
		// If *something* happened, then reload things and change the text, too...
		reloadEverything(); // <-- this also makes the settings/markers "clean"
		textDisplayInfo->setText("Settings have been applied.\nThis does not save the changes to disk, however.");
	}
}

bool SettingsDialog::markersChanged() {
	return markersChangedAtAll;
}

bool SettingsDialog::settingsChanged() {
	return settingsChangedAtAll;
}

void SettingsDialog::discardChanges() {
	
	switch(QMessageBox::warning(this, "GenomeMixer: Discard Changes?",
								"Should we really revert to the previous settings?\n" 
								"You will lose any changes made since you switched to this tab.",
								"Cancel", "Discard changes!",
								0,
								1)) {
		case 0: // cancel
		textDisplayInfo->setText("\"Revert\" operation cancelled.\n");
		break;
	case 1:
		reloadEverything(); // reload from the last apply...
		textDisplayInfo->setText("Changes were discarded.\nNow using the last-applied settings.");
		makeAllClean();
		break;
	}
}

// ==============================================================================================
// ##############################################################################
// ##################################### CHROMOSOME FUNCTIONS #################
// ################################################################################

void SettingsDialog::chrTableApply() {
	// Take the data from the user-visible table, and save it into the internal data structures.
	// Note that this doesn't save anything to *disk*.    
	
	Chromosome::clearAllChrData(); // <-- resets the number of chromosomes to 0 as well
	Chromosome::setNumberOfChromosomes(tableChr->numRows() + 2); // <-- add two for the sex chrs
	
	// "i" has scope beyond the for loop: it is ued in the two lines below, as well
	int i = 0;
	while (i < tableChr->numRows()) {
		Chromosome::addChrData(QString(QString("Chr_") + QString::number(i+1)),
							   atof(tableChr->text(i, chr_MB_col)),
							   atof(tableChr->text(i, chr_cM_col)));
		i++;
	}
	Chromosome::addChrData(QString(QString("Chr_X_") + QString::number(i+1)),
						   atof(tableSexChr->text(0, chr_MB_col)),
						   atof(tableSexChr->text(0, chr_cM_col))  );
	Chromosome::addChrData(QString(QString("Chr_Y_") + QString::number(i+2)),
						   atof(tableSexChr->text(1, chr_MB_col)),
						   atof(tableSexChr->text(1, chr_cM_col))  );
}



void SettingsDialog::reloadChrData() {
	assert(tableChr->numCols() == 2); // two columns total: one for the MB, one for the cM
	
	int numAutosomes = std::max(0, Chromosome::numChr() - 2);
	tableChr->setNumRows(numAutosomes);
	
	tableChr->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	for (int row = 0; row < numAutosomes; row++) {
		assert(row < tableChr->numRows());
		assert(row < Chromosome::numChr());
		tableChr->setText(row, chr_MB_col, QString::number(Chromosome::lengthMBASEfromIndex(row)));
		tableChr->setText(row, chr_cM_col, QString::number(Chromosome::length_cM_fromIndex(row)));	
		tableChr->setRowStretchable(row, false);
	}
	
	if (Chromosome::numChr() >= 2) {
		// Set the X and Y chromosome data
		assert(Chromosome::xChrIndex() < Chromosome::numChr());
		assert(Chromosome::yChrIndex() < Chromosome::numChr());
		assert(tableSexChr->numRows() == 2);
		
		tableSexChr->setText(0, chr_MB_col, QString::number(Chromosome::lengthMBASEfromIndex(Chromosome::xChrIndex())));
		tableSexChr->setText(0, chr_cM_col, QString::number(Chromosome::length_cM_fromIndex(Chromosome::xChrIndex())));
		tableSexChr->setText(1, chr_MB_col, QString::number(Chromosome::lengthMBASEfromIndex(Chromosome::yChrIndex())));
		tableSexChr->setText(1, chr_cM_col, QString::number(Chromosome::length_cM_fromIndex(Chromosome::yChrIndex())));
	} else {
		// Set the fields to 0
		tableSexChr->setText(0, chr_MB_col, "0.0");
		tableSexChr->setText(0, chr_cM_col,  "0.0");
		tableSexChr->setText(1, chr_MB_col, "0.0");
		tableSexChr->setText(1, chr_cM_col,  "0.0");
	}
}


void SettingsDialog::updateChromosomeInfo() {
	// Updates the total number of chromosomes
	textNumChr->setText("Number of Chromosomes: "
						+ QString::number(tableChr->numRows() + 2));
	
	// Updates the total megabases & cM
	double total_MB = 0, total_cM = 0;
	for (int row = 0; row < tableChr->numRows(); row++) {
		// Count up the autosome length
		total_MB += atof(tableChr->text(row, chr_MB_col));
		total_cM += atof(tableChr->text(row, chr_cM_col));
	}
	
	for (int sexChrRow = 0; sexChrRow < tableSexChr->numRows(); sexChrRow++) {
		// Count up the X and Y chromosomes' lengths
		total_MB += atof(tableSexChr->text(sexChrRow, chr_MB_col));
		total_cM += atof(tableSexChr->text(sexChrRow, chr_cM_col));
	}
	
	textMBValue->setText(QString::number(total_MB));
	textCMValue->setText(QString::number(total_cM));
	
	// Set the part of the text that describes how 
	if (mainMixPtr->noRecombDistBases == 0) {
		textInterference->setText(QString("No interference; recombinations can\noccur anywhere on the chromosome."));
	} else {
		textInterference->setText(QString("Absolute interference, with at least\n") + QString::number(double(mainMixPtr->noRecombDistBases / 1000000.0)) + " MBases between recombinations.");
	}
	
	switch (mainMixPtr->recombModel) {
	case kNO_RECOMB:
		textRecomb->setText(QString("Recombination is off; no recombination will occur."));
		break;
	case kLIN_RECOMB:
		textRecomb->setText(QString("Linear recombination equation."));
		break;
	case kHALDANE_RECOMB:
		textRecomb->setText(QString("Haldane recombination equation."));
		break;
	default:
		textRecomb->setText(QString("Unknown recombination model: programming error?"));
		break;
	}
	
	updateSaveDiscardButtons();
}




void SettingsDialog::chrNew() {
	int newIndex = max(tableChr->currentRow() + 1, 0); // add a new marker after the selected one
	
	tableChr->insertRows(newIndex);
	QString chrNumStr = QString::number(newIndex+1);
	tableChr->verticalHeader()->setLabel(newIndex, QIconSet(NULL), chrNumStr);
	
	tableChr->ensureCellVisible(newIndex, max(0, tableChr->currentColumn()) ); // scrolls to the new chromosome
	tableChr->setCurrentCell(newIndex, max(0, tableChr->currentColumn()) );
	
	tableChr->setText(newIndex, chr_MB_col, "0.0");
	tableChr->setText(newIndex, chr_cM_col, "0.0");
	
	renumberTable(tableChr, newIndex, tableChr->numRows()); // RE-LABEL
	tableChr->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
	makeCurrentPaneDirty();
	updateChromosomeInfo();
}


void SettingsDialog::chrDelete() {
	int rowToRemove = tableChr->currentRow();
	if (rowToRemove < 0) return;
	
	if (tableChr->numRows() <= 0 || rowToRemove < 0 || rowToRemove >= tableChr->numRows()) {
		QMessageBox::warning(this, "GenomeMixer Alert",
							 "You must have a chromosome selected in\n"
							 "order for the delete button to have any effect.\n"
							 "Also, you can't delete the X or Y chromosomes.",
							 "Ok");
		return;
	}
	assert(rowToRemove >= 0 && rowToRemove < tableChr->numRows());
	
	tableChr->removeRow(rowToRemove);
	tableChr->ensureCellVisible(rowToRemove, tableChr->currentColumn()); // scrolls to the new chromosome
	tableChr->setCurrentCell(rowToRemove, tableChr->currentColumn()); // set the current cell...
	renumberTable(tableChr, rowToRemove, tableChr->numRows()); // RE-LABEL
	
	tableChr->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
	makeCurrentPaneDirty();
	updateChromosomeInfo();
}




// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// ##########################################################################
// ##################################### MARKER   FUNCTIONS ################
// ##########################################################################

void SettingsDialog::markersTableApply() {
	// Save the user-visible marker table into the internal data structure, overwriting whatever was in that structure already (if anything)
	if (mainMixPtr->markerColPtr) {
		delete mainMixPtr->markerColPtr; // <-- regular delete, NOT delete[ ] (that seems to cause a segfault)
		mainMixPtr->markerColPtr = NULL;
	}
	mainMixPtr->markerColPtr = new MarkerCollection(tableMarkers->numRows());
	
	for (int i = 0; i < tableMarkers->numRows(); i++) {
		QString chrID = tableMarkers->text(i, markers_chr_col);
		int whichChrIndexToAddTo = 0;
		if (chrID.startsWith(QString("X")) || chrID.startsWith(QString("x"))) {
			QMessageBox::warning(this, "GenomeMixer Alert",
								 "Sorry, but we don't currently know how to convert from\nX to a chromosome number. Instead, just make\nthe X Chr the next-to-last chromosome, and the Y Chr\nthe very last chromosome. In mice (21 chromosomes), X would be 20 and Y would be 21.", "Ok");
			//whichChrIndexToAddTo = Chromosome::xChrIndex();
			
		} else if (chrID.startsWith(QString("Y")) || chrID.startsWith(QString("y"))) {
			QMessageBox::warning(this, "GenomeMixer Alert",
								 "Sorry, but we don't currently know how to convert from\nY to a chromosome number. Instead, just make\nthe Y Chr the last chromosome, and the X Chr\nthe next-to-last chromosome. In mice (21 chromosomes), X would be 20 and Y would be 21.", "Ok");
			//whichChrIndexToAddTo = Chromosome::yChrIndex();
			
		} else {
			whichChrIndexToAddTo = std::max(0, int(atoi(chrID) - 1)); // <-- hopefully, it is a number in this case
			// note that we subtract one: because if we display chromosome "1," that is really at index 0.
			// But we also make sure to cap it off at 0; we don't want any "Chr -1"s.
		}
		//qDebug("Adding to this chromosome: " + QString::number(whichChrIndexToAddTo) + QString(": ") + chrID);
		
		if (whichChrIndexToAddTo >= 0) {
			mainMixPtr->markerColPtr->addMarker(whichChrIndexToAddTo,
												tableMarkers->text(i, markers_name_col),
												atol(tableMarkers->text(i, markers_bases_col)),
												atof(tableMarkers->text(i, markers_cM_col))  );
		} else {
			QMessageBox::warning(this, "GenomeMixer Alert",
								 "A chromosome is not valid,\n"
								 "so we can't add a marker there.\n"
								 "If you want to add a chromosome at that location,\n"
								 "you should edit the chromosome settings (in the chromosome tab).",
								 "Ok");
		}
	}
}



void SettingsDialog::reloadMarkerData() {	
	assert(tableMarkers->numCols() == 4); //  four columns total, or else something is not right
	MarkerCollection *markers = mainMixPtr->markerColPtr;
	if (markers == NULL || markers->totalMarkers() <= 0) {
		// In other words, if the markers still haven't been initialized!
		tableMarkers->setNumRows(0);
		tableMarkers->updateHeaderStates();
		return;
	}
	
	tableMarkers->setNumRows(markers->totalMarkers());
	tableMarkers->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
	// Here's how this works: Markers are stored by chromosome and by marker *on that chromosme*.
	// However, we want "row" to increment from 0 to the total number of markers, so we keep it outside the loop.
	// For each chromosome, we go from 0 to the number of markers on that chromsome, resetting to 0 with each chromosome,
	// but we *don't* reset "row"--that just keeps incrementing.
	int row = 0;
	for (int chrIndex = 0;   chrIndex < markers->numberOfChrs();   chrIndex++) {
		
		QString chrNumberStr = QString::number(chrIndex+1);
		/*if (chrIndex == Chromosome::xChrIndex()) {
		chrNumberStr.prepend("X: ");
	} else if (chrIndex == Chromosome::yChrIndex()) {
		chrNumberStr.prepend("Y: ");
	}*/
		
		for (long markerIndex = 0;   markerIndex < markers->markersInChr(chrIndex);   markerIndex++) {
			
			tableMarkers->setText(row, markers_chr_col, chrNumberStr);
			tableMarkers->setText(row, markers_name_col,  markers->name(chrIndex, markerIndex));
			tableMarkers->setText(row, markers_bases_col, QString::number(markers->base(chrIndex, markerIndex)) );
			tableMarkers->setText(row, markers_cM_col,     QString::number(markers->cM(chrIndex, markerIndex)) );
			row++;
		}
	}
}



void SettingsDialog::updateMarkerInfo() {
	// Just sets the little text label that tells us the total number of markers.
	QString numberOfMarkersStr= QString::number(tableMarkers->numRows());
	textMarkersTotal->setText("Total Markers: " + numberOfMarkersStr);
}



void SettingsDialog::markerNew() {
	int newIndex = max(tableMarkers->currentRow() + 1, 0); // add a new marker after the selected one
	tableMarkers->insertRows(newIndex);
	tableMarkers->ensureCellVisible(newIndex, max(0, tableMarkers->currentColumn()) ); // scrolls to the new marker
	tableMarkers->setCurrentCell(newIndex, max(0, tableMarkers->currentColumn()) );
	
	tableMarkers->setText(newIndex, markers_name_col, "NewMarker");
	tableMarkers->setText(newIndex, markers_chr_col, "none");
	tableMarkers->setText(newIndex, markers_bases_col, "none");
	tableMarkers->setText(newIndex, markers_cM_col, "none");
	
	renumberTable(tableMarkers, newIndex, tableMarkers->numRows()); // RE-LABEL
	
	tableMarkers->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
	makeCurrentPaneDirty();
	updateMarkerInfo();
}



void SettingsDialog::markerDelete() {
	int rowToRemove = tableMarkers->currentRow();
	if (rowToRemove < 0) return;
	
	if (rowToRemove < 0 || rowToRemove >= tableMarkers->numRows()) {
		QMessageBox::warning(this, "GenomeMixer Alert",
							 "You don't have any marker selected to delete.\n"
							 "Consequently, the delete button has no effect",
							 "That makes sense");
		return;
	}
	tableMarkers->removeRow(rowToRemove);    
	renumberTable(tableMarkers, rowToRemove, tableMarkers->numRows()); // RE-LABEL
	
	tableMarkers->setCurrentCell(rowToRemove, tableMarkers->currentColumn()); // set the current cell...
	tableMarkers->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
	makeCurrentPaneDirty();
	updateMarkerInfo();
}



void SettingsDialog::chooseInterferenceModel() {
	InterferenceDialog *interDialog = new InterferenceDialog();
	
	// set the current model, based on our current GenomeMixer's settings
	if (mainMixPtr->noRecombDistBases == 0) {
		interDialog->radioNoInterfere->setChecked(true);
		interDialog->absInterfereEntry->setText(QString("25.0"));
	} else {
		interDialog->radioAbsolute->setChecked(true);
		interDialog->absInterfereEntry->setText(QString::number(double(mainMixPtr->noRecombDistBases / 1000000.0)));
	}
		
	if (interDialog->exec() == QDialog::Accepted) {
		
		if (interDialog->radioNoInterfere->isChecked()) {
			// NO INTERFERENCE AT ALL
			mainMixPtr->noRecombDistBases = 0;
			// END of no intereference at all
			
		} else if (interDialog->radioAbsolute->isChecked()) {
			// ABSOLUTE INTERFERENCE
			double distInMB = atof(interDialog->absInterfereEntry->text());
			if (distInMB < 0) {
				distInMB = -distInMB; // if it's less than 0, just flip it
				interDialog->absInterfereEntry->setText(QString::number(distInMB));
			}
			if (distInMB > 2000) {
				cerr << "2000 megabases between recombinations?" << endl;
				QMessageBox::warning(this, "GenomeMixer Warning",
									 "Sorry, but you can't require more\n" 
									 "space between recombinations than 2000\n"
									 "megabases. Note that 2000 MB is 2,000,000,000 bases,\n"
									 "which would be a really enormous chromosome. Maybe\n"
									 "you entered the number in bases instead of MB?",
									 "Ok");
				distInMB = 2000; // maximum...
				interDialog->absInterfereEntry->setText(QString::number(distInMB));
			}
			// A long can be at least 2 billion, so 2000 (the max of distInMB) times 1 million is OK
			mainMixPtr->noRecombDistBases = long(distInMB * 1000000.0);
			// END of absolute interference
		}
		
		makeCurrentPaneDirty();
		updateChromosomeInfo();
	}
	
	if (interDialog) delete interDialog; // free it after we are done...			
}


void SettingsDialog::chooseRecombinationModel() {
	
	RecombinationDialog *recombDialog = new RecombinationDialog();
	
	// set the current model, based on our current GenomeMixer's settings
	switch (mainMixPtr->recombModel) {
	case kNO_RECOMB:
		recombDialog->radioNoRecomb->setChecked(true);
		break;
	case kLIN_RECOMB:
		recombDialog->radioLinear->setChecked(true);
		break;
	default:
		// ???
		break;
	}
	
	if (recombDialog->exec() == QDialog::Accepted) {		
		if (recombDialog->radioNoRecomb->isChecked()) {
			// NO RECOMBINATION AT ALL
			mainMixPtr->recombModel = kNO_RECOMB;
			
		} else if (recombDialog->radioLinear->isChecked()) {
			// LINEAR RECOMBINATION MODEL
			mainMixPtr->recombModel = kLIN_RECOMB;
		}		
		makeCurrentPaneDirty();
		updateChromosomeInfo();
	}
	
	if (recombDialog) delete recombDialog; // free it after we are done...			
}


void SettingsDialog::markersAutoGenerate() {
	AutoGenMarkers *genDialog = new AutoGenMarkers();
	
	if (genDialog->exec() == QDialog::Accepted) {
		genDialog->highlightTextEntry();
		
		//int typeOfSpacing = 0;
		// switch (typeOfSpacing)
		
		float distInCentimorgans = atof(genDialog->editMarkerSpacing->text());
		
		if (distInCentimorgans < 0.0001) {
			QMessageBox::warning(this, "GenomeMixer Warning",
								 "Sorry, but you can't space markers\n" 
								 "any closer than 0.0001 cM.",
								 "Darn");
			return;
		} else if (distInCentimorgans < 0.04999) {
			switch(QMessageBox::warning(this, "GenomeMixer Warning",
										"Are you sure you want to generate markers\n" 
										"spaced so close together? This may end up\n"
										"making so many markers that GenomeMixer will\n"
										"become incredibly slow.",
										"Never mind", "Make them anyway",
										0,
										1)) {
			case 0: // cancel
				return; // I guess the user doesn't want a zillion markers after all
				
				break;
			case 1:
				// "make all the markers anyway, I don't care how many there are"
				break;
			}

		}
			
		int newNumberOfMarkers = int(ceil(Chromosome::allChrTotal_cM() / distInCentimorgans)) + 2*Chromosome::numChr() + 1;
		// This is the maximum number of markers we could have total, no matter what the split in cM among chromosomes was	    
		
		tableMarkers->setNumRows(newNumberOfMarkers); // we will need at least this many rows, but probably fewer (adjust to the proper number at the end)	        
		int row = 0;	    
		for (int chrIndex = 0;   chrIndex < Chromosome::numChr();   chrIndex++) {
				QString chrNumberStr = QString::number(chrIndex+1);		
				int markerIndexThisChr = 0;
				bool doneWithChr = false;
				for (float cM_locThisChr = 0.0; !doneWithChr; cM_locThisChr += distInCentimorgans) {
					
					if (cM_locThisChr >= Chromosome::length_cM_fromIndex(chrIndex)) {
						// This means we either got to or exceeded the length of the chromosome,
						// so we should add the "end cap" marker onto the chromosome, and then we are done with it.
						cM_locThisChr = Chromosome::length_cM_fromIndex(chrIndex);
						doneWithChr = true;
					}			    
					assert(row < tableMarkers->numRows());
					tableMarkers->setText(row, markers_chr_col, chrNumberStr);
					tableMarkers->setText(row, markers_name_col,  QString("C")
										  + chrNumberStr + QString("_Mark")
										  + QString::number(markerIndexThisChr+1)
										  + (doneWithChr ?                     QString("_end") : QString::null)
										  + (markerIndexThisChr == 0 ?    QString("_start") : QString::null)
										  );
					tableMarkers->setText(row, markers_bases_col, QString::number(long( float(cM_locThisChr/Chromosome::length_cM_fromIndex(chrIndex)) * Chromosome::lengthBasesFromIndex(chrIndex)) ));
					tableMarkers->setText(row, markers_cM_col, QString::number(cM_locThisChr));
					row++;
					markerIndexThisChr++;
				}
		}			
		tableMarkers->setNumRows(row); // <-- now we set it to the actual number of rows
		tableMarkers->updateHeaderStates(); // <-- needs to be called after removing/adding cells
		
		makeCurrentPaneDirty();
		updateMarkerInfo();
	}
	
	
	if (genDialog) delete genDialog; // free it after we are done...	
}


// ***************************************************************************************************************
// ***************************************************************************************************************
// ##########################################################################
// ##################################### PUREBRED FUNCTIONS ###############
// ##########################################################################

void SettingsDialog::purebredTableApply() {    
	// Saves all the purebred table data from the user-visible tables into the not-visible data structures.
	// Doesn't save anything to disk, though.
	mainMixPtr->clearAllParentalStrains();
	for (int i = 0; i < tablePurebreds->numRows(); i++) {
		mainMixPtr->purePtr()->addPurebred(tablePurebreds->text(i, purebreds_name_col),
										   tablePurebreds->text(i, purebreds_symbol_col));
	}
}

void SettingsDialog::validatePurebredItem(int row, int col) {
	if (row < 0 || col < 0 || row >= tablePurebreds->numRows() || col >= tablePurebreds->numCols()) {
		updatePurebredInfo();
		return; // out of bounds, so nothing to check...
	}
	QString cellText = tablePurebreds->text(row, col);
	
	if (col == purebreds_name_col) {
		// Make sure it doesn't have any commas, spaces, or slashes in the name
		if (cellText.isNull() || cellText.contains(' ', true)
			|| cellText.contains(',', true) || cellText.contains('\"', true) || cellText.contains('\'', true)) {
			// So... either this cell is blank, or it contains a Forbidden Symbol
			tablePurebreds->setText(row, col, QString("Purebred_") + QString::number(2000 + row));
			
			QMessageBox::warning(this, "GenomeMixer Alert",
								 "The purebred name you just entered is invalid.\n"
								 "Purebred strain names cannot contain spaces,\n"
								 "commas, or quotation marks. Also, they cannot\n"
								 "be blank. If you want to delete a purebred strain,\n"
								 "select it and press the delete button instead.",
								 "Ok");
		}
		
	} else if (col == purebreds_symbol_col) {
		bool ok = true; // <-- not actually used, but required for "toInt"
		QRegExp checkSymbolExp("^([A-Za-z]|[0-9]+)$", false);
		// Check to make sure it's either a single letter, or it's a number from 0 to 30,000
		if (cellText.isNull() || !(cellText.contains(checkSymbolExp)) || ((cellText.toInt(&ok, 10) > 30000) && ok)) {
			tablePurebreds->setText(row, col, QString::number(2000 + row));
			QMessageBox::warning(this, "GenomeMixer Alert",
								 "The purebred symbol you just entered is invalid.\n"
								 "Purebred symbols must either be a single capital or\n"
								 "lower-case letter (A-Z or a-z), or a number from 0\n"
								 "to 30,000, inclusive. They calso cannot be blank.\n"
								 "If you want to delete a purebred strain, select it\n"
								 "and press the delete button instead.",
								 "Ok");
		}
	} else {
		// we shouldn't get here
		cerr << "Calling SettingsDialog::validatePurebredItem on an invalid item. Row was " << row << ", and column was " << col << endl;
	}
	
	makeCurrentPaneDirty();
	updatePurebredInfo();
}

void SettingsDialog::reloadPurebredData() {    
	assert(mainMixPtr != NULL);
	assert(tablePurebreds->numCols() == 2); // two columns total: one for the Name, one for the Symbol
	
	tablePurebreds->setNumRows(mainMixPtr->numParentalStrains());
	tablePurebreds->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
	for (int row = 0; row < mainMixPtr->numParentalStrains(); row++) {
		assert(row < tablePurebreds->numRows());
		tablePurebreds->setText(row, purebreds_name_col, mainMixPtr->getPurebredName(row));
		tablePurebreds->setText(row, purebreds_symbol_col, mainMixPtr->getPurebredSymbol(row));	
	}
}

void SettingsDialog::updatePurebredInfo() {
	textPurebredNum->setText(QString("Number of Purebred Strains: ") + QString::number(tablePurebreds->numRows()));
}



void SettingsDialog::purebredNew() {
	int newIndex = max(tablePurebreds->currentRow() + 1, 0); // add a new marker after the selected one
	tablePurebreds->insertRows(newIndex);
	tablePurebreds->ensureCellVisible(newIndex, max(0, tablePurebreds->currentColumn()) ); // scrolls to the new chromosome
	tablePurebreds->setCurrentCell(newIndex, max(0, tablePurebreds->currentColumn()) );
	
	tablePurebreds->setText(newIndex, purebreds_name_col, QString("PurebredStrain_") + QString::number(newIndex));
	tablePurebreds->setText(newIndex, purebreds_symbol_col, QString::number(1000 + newIndex));
	
	renumberTable(tablePurebreds, newIndex, tablePurebreds->numRows()); // RE-LABEL
	
	tablePurebreds->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
	makeCurrentPaneDirty();
	updatePurebredInfo();
}



void SettingsDialog::purebredDelete() {
	int rowToRemove = tablePurebreds->currentRow();
	if (rowToRemove < 0) return;
	
	if (rowToRemove < 0 || rowToRemove >= tablePurebreds->numRows()) {
		QMessageBox::warning(this, "GenomeMixer Alert",
							 "You don't have any purebred strain selected to delete.",
							 "Ok");
		return;
	}
	
	tablePurebreds->removeRow(rowToRemove);
	tablePurebreds->setCurrentCell(rowToRemove, tablePurebreds->currentColumn()); // set the current cell...
	
	renumberTable(tablePurebreds, rowToRemove, tablePurebreds->numRows()); // RE-LABEL
	
	tablePurebreds->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
	makeCurrentPaneDirty();
	updatePurebredInfo();
}


// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// ##########################################################################
// ############################# BREEDING SCHEME FUNCTIONS ###############
// ###########################################################################

void SettingsDialog::schemeTableApply() {
	assert(mainMixPtr != NULL);
	
	mainMixPtr->setTableSize(tableScheme->numRows(), tableScheme->numCols());
	for (int cage = 0; cage < tableScheme->numRows(); cage++) {
		for (int gen = 0; gen < tableScheme->numCols(); gen++) {
			// Takes the changes from tableEntries and applies them to the mainMixPtr
			mainMixPtr->setCellData(cage, gen, tableEntries[cage][gen]);
		}
	}
}

void SettingsDialog::reloadSchemeData(void) {
	assert(mainMixPtr != NULL);
	
	tableScheme->setNumRows(mainMixPtr->numCages());
	tableScheme->setNumCols(mainMixPtr->numGenerations());
	spinSchemeCages->setValue(tableScheme->numRows()); // set the spinners properly
	spinSchemeGenerations->setValue(tableScheme->numCols()); // set the spinners properly
	
	renumberCages(tableScheme, 0, tableScheme->numRows());
	renumberGenerations(tableScheme, 0, tableScheme->numCols());
	
	tableScheme->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	
	for (int row = 0; row < tableScheme->numRows(); row++) {
		for (int col = 0; col < tableScheme->numCols(); col++) {
			
			if (tableEntries[row][col]) {
				// If there is already something in our table of copies, delete it
				delete (tableEntries[row][col]);
				tableEntries[row][col] = NULL;
			}
			
			// And copy from the main table into this tableEntries table (a table of pointers to CellData)	  
			if (mainMixPtr->getCellData(row, col) != NULL) {
				tableEntries[row][col] = new CellData(mainMixPtr->getCellData(row, col));
			}
			
			
			// Now, set the actual text cells...
			if (tableEntries[row][col] != NULL) {
				// If there's something there, fill the cell accordingly
				tableScheme->setText(row, col, (tableEntries[row][col])->toString(mainMixPtr->purePtr()));		
			} else {
				// Otherwise, it's a blank cell
				tableScheme->setText(row, col, QString::null);
			}
			
			tableScheme->adjustColumn(col); // Auto-resize the column to fit the data
		}
	}
}


void SettingsDialog::currentCellHasChanged(int row, int col) {
	static int previousRow = 0; // rows are cages
	static int previousCol = 0; // columns are generations
	
	if (tableEntries[row][col] == NULL) {
		if (previousRow >= 0 && previousCol >= 0 && previousRow <= kTABLE_SIZE && previousCol <= kTABLE_SIZE
			&& previousCol < col) {
			// above line: if (previousCol < col) is there so that we won't try to forward-reference generations (say you click on generation 20, and then 15: 15 can't have 20 as a parent, so it just shouldn't do anything at all
			if (tableEntries[previousRow][previousCol] != NULL) {
				// Make this new cell have its parents be the last-clicked cell, and have the same number off
				// mice made, as well
				tableEntries[row][col] = new CellData(tableEntries[previousRow][previousCol]);
				tableEntries[row][col]->setMother(previousRow, previousCol); // cage, gen
				tableEntries[row][col]->setFather(previousRow, previousCol); // cage, gen
			}
		}
	}
	
	previousRow = row; // update 'em
	previousCol = col; // update 'em
	
	updateSchemeInfo(row, col);
}


void SettingsDialog::updateSchemeInfo(void) { // (overloaded function)
	// Overloaded: just calls updateSchemeInfo(...) with the current cell
	updateSchemeInfo(tableScheme->currentRow(), tableScheme->currentColumn());
}

void SettingsDialog::updateSchemeInfo(int row, int col) {
	// DON'T check row, col for invalidness here: we do that below, and adjust the controls accordingly
	
	// Check to see if the cell is commented out (with a "#") (note that this is an overloaded function)
	
	bool validSelection = (row >= 0 && col >= 0) && (tableScheme->numRows() > 0 && tableScheme->numCols() > 0);
	static bool lastValidSelection = !validSelection; // keeps track of changes in validity, so we know if we have to enable/disable everything
	// note that lastValidSelection is *static*: it remembers what its value was from last time, so it only gets initialized here once
	
	if (!validSelection) {
		// The user has somehow selected a cell that is out of the bounds of the table,
		// or possibly resized the table so that it has 0 rows or 0 columns
		parentsGroup->setTitle("(No Cell Selected)");
		progenyGroup->setTitle("(No Cell Selected)");
		optionsGroup->setTitle("(No Cell Selected)");	
		schemeCheckEnabled->setChecked(false);
		schemeCheckExport->setChecked(false);
		schemeParentFem->setText(QString::null);
		schemeParentMale->setText(QString::null);
		
	} else if (!tableEntries[row][col]) {
		// The user has selected a cell, but it's a NULL cell right now
		parentsGroup->setTitle("(Empty cell)");
		progenyGroup->setTitle("(Empty cell)");
		optionsGroup->setTitle("(Empty cell)");
		schemeCheckEnabled->setChecked(false);
		schemeCheckExport->setChecked(false);
		schemeParentFem->setText(QString::null);
		schemeParentMale->setText(QString::null);
		
	} else {
		assert(row >= 0 && col >= 0 && row < kTABLE_SIZE && col < kTABLE_SIZE);
		assert(tableEntries[row][col]);
		
		tableScheme->setText(row, col, tableEntries[row][col]->toString(mainMixPtr->purePtr())); // set the actual user-visible text of the cell
		
		schemeCheckExport->setChecked(tableEntries[row][col]->toBeExported());
		
		this->schemeSpinRandomSex->setValue(tableEntries[row][col]->randomSex()); // update the "total mice" spinner
		this->schemeSpinFem->setValue(tableEntries[row][col]->minFemales());// update the "minimum number of females" spinner
		this->schemeSpinMale->setValue(tableEntries[row][col]->minMales()); // update the "minimum number of males" spinner
		
		schemeParentFem->setText(tableEntries[row][col]->motherString(mainMixPtr->purePtr())); // <-- set the text for the mother
		// Indicate whether the mother is a purebred, or if it is from one of the cells
		if (tableEntries[row][col]->motherIsPurebred()) {
			schemePulldownFem->setCurrentItem(pulldown_purebred);
		} else {
			schemePulldownFem->setCurrentItem(pulldown_from_cell);
		}
		
		schemeParentMale->setText(tableEntries[row][col]->fatherString(mainMixPtr->purePtr())); // <-- set the text for the father
		// Indicate whether the father is a purebred, or if it is from one of the cells
		if (tableEntries[row][col]->fatherIsPurebred()) {
			schemePulldownMale->setCurrentItem(pulldown_purebred);
		} else {
			schemePulldownMale->setCurrentItem(pulldown_from_cell);
		}
		
		QString nameStr = QString(lettersFromNum(col) + QString::number(row + 1)); // <-- this will be something like "A51" or "BC93"
		
		if ((tableEntries[row][col])->commentedOut()) {
			schemeCheckEnabled->setChecked(false); // it's NOT enabled--uncheck the checkbox
			optionsGroup->setTitle("Options for " + nameStr + " (inactive)");
		} else {
			schemeCheckEnabled->setChecked(true); // it IS enabled--check the checkbox
			optionsGroup->setTitle("Options for " + nameStr);
		}	
		parentsGroup->setTitle("Parents for " + nameStr);
		progenyGroup->setTitle("Progeny for " + nameStr);
	}
	
	if (lastValidSelection != validSelection) {
		// In other words, the selection changed from valid to invalid, or vice-versa
		schemePulldownFem->setEnabled(validSelection);
		schemePulldownMale->setEnabled(validSelection);
		schemeCheckEnabled->setEnabled(validSelection);
		schemeCheckExport->setEnabled(validSelection);
		schemeParentFem->setEnabled(validSelection);
		schemeParentMale->setEnabled(validSelection);
		schemeSpinRandomSex->setEnabled(validSelection);
		schemeSpinFem->setEnabled(validSelection);
		schemeSpinMale->setEnabled(validSelection);
	}
	
	if (col >= 0 && col < tableScheme->numCols())
		tableScheme->adjustColumn(col); // if this is a real column, then we'll resize it
	
	lastValidSelection = validSelection;
}


QString SettingsDialog::lettersFromNum(int num) {
	assert(num >= 0);
	// returns "A" from 0, "Z" from 25, "AA" from 26, "AB" from 27, etc.
	if (num <= 25) {
		return (QString(QChar('A' + num))); // <-- 0 == 'A', 25 == 'Z'
	} else if (num >= 26) {
		return (QString( QChar('A' + int(num / 26) - 1 )) + QString(QChar('A' + num % 26))); // 27 == "AA", 28 == "AB", etc.
	} else {
		return (QString(">=AAA"));
	}
}


void SettingsDialog::renumberCages(QTable *table, int start, int finishNonInclusive) {
	assert(start >= 0 && finishNonInclusive <= table->numRows() && start <= finishNonInclusive);
	for (int i = start; i < finishNonInclusive; i++) {
		// Set the labels to 1, 2, 3, ... 
		table->verticalHeader()->setLabel(i, QIconSet(NULL), "Cage (" + QString::number(i+1) + ")"); // Re-number
	}
}


void SettingsDialog::renumberGenerations(QTable *table, int start, int finishNonInclusive) {
	assert(start >= 0 && finishNonInclusive <= table->numCols() && start <= finishNonInclusive);
	for (int i = start; i < finishNonInclusive; i++) {
		// Set the labels to 1, 2, 3, ... 
		QString lettersStr = lettersFromNum(i); // <-- like "A" or "AZ"--gives the generation Excel-style letters in the header
		table->horizontalHeader()->setLabel(i, QIconSet(NULL), QString("(") + lettersStr + QString(") Gen ") + QString::number(i+1)); // Re-number
	}
	
}

void SettingsDialog::schemeChangeNumCages( int newCages ) {
	const int oldNumCages = tableScheme->numRows();
	
	if (newCages < 0) {
		QMessageBox::warning(this, "GenomeMixer Alert",
							 "You can't have a negative number of cages.",
							 "I agree");
		return;
	} else if (newCages > kTABLE_SIZE) {  // >, not >=
		QMessageBox::warning(this, "GenomeMixer Alert",
							 "Sorry, we can't handle this many cages.",
							 "Ok");
		return;
	}
	
	if (newCages > oldNumCages) {
		// have to add a new blank row
		tableScheme->setNumRows(newCages);
		renumberCages(tableScheme, oldNumCages, newCages);
		
	} else if (newCages < oldNumCages) {
		// have to remove some cages (rows) in an intelligent manner
		tableScheme->setNumRows(newCages);
		
	} else {
		return; // hmm, they are equal, so nothing changes
	}
	tableScheme->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	makeCurrentPaneDirty();
	updateSchemeInfo();
}


void SettingsDialog::schemeChangeNumGen( int newGens ) {
	const int oldNumGens = tableScheme->numCols();
	
	if (newGens < 0) {
		QMessageBox::warning(this, "GenomeMixer Alert", 
							 "You can't have a negative number of generations.",
							 "Ok");
		return;
	} else if (newGens > kTABLE_SIZE) { // >, not >=
		QMessageBox::warning(this, "GenomeMixer Alert", 
							 "Sorry, we can't handle this many generations.\n"
							 "Ok");
		return;
	}
	
	if (newGens > oldNumGens) {
		tableScheme->setNumCols(newGens);
		renumberGenerations(tableScheme, oldNumGens, newGens);
		
	} else if (newGens < oldNumGens) {
		tableScheme->setNumCols(newGens);
		
	} else {
		return; // hmm, they are equal, so nothing changes
	}
	tableScheme->updateHeaderStates(); // <-- needs to be called after removing/adding cells
	makeCurrentPaneDirty();
	updateSchemeInfo();
}


void SettingsDialog::makeCellWhenThereIsNone(int row, int col) {
	if (!tableEntries[row][col]) {  // Make a new CellData if there isn't one
		tableEntries[row][col] = new CellData();
	}
}


void SettingsDialog::schemeChangeEnabled( bool state ) {
	// Changes between commented/uncommented
	int row = tableScheme->currentRow();
	int col = tableScheme->currentColumn();
	if (row < 0 || col < 0) return;
	
	assert(row < kTABLE_SIZE && col < kTABLE_SIZE);    
	
	makeCellWhenThereIsNone(row, col);
	
	if (state == !tableEntries[row][col]->commentedOut()) {
		// state = "is it enabled?"
		// commented out = "is it commented out = not enabled?"
		// so if state != comment, that means they were the same already, so we shouldn't do anything
		return;
	} else {
		// Ok... so we do need to change the "enabled" state
		tableEntries[row][col]->setCommentedOut(!state); // <-- commented = the opposite of the "enabled" state     	   
		makeCurrentPaneDirty(); // we changed the "enabled" status
		updateSchemeInfo();
	}
}

void SettingsDialog::schemeChangeExport( bool state ) {
	// Deals with the "exp" at the beginning of a string, that indicates whether or not to export it
	int row = tableScheme->currentRow();
	int col = tableScheme->currentColumn();
	if (row < 0 || col < 0) return;
	
	assert(row >= 0 && col >= 0 && row < kTABLE_SIZE && col < kTABLE_SIZE);
	
	this->makeCellWhenThereIsNone(row, col);
	
	if (state == tableEntries[row][col]->toBeExported()) {
		// if they are already in agreement, then we don't need to do anything
		return;
	} else {
		tableEntries[row][col]->setToBeExported(state); // cell should be exported now
		makeCurrentPaneDirty(); // we changed the "exported" status
		updateSchemeInfo();
	}
}


void SettingsDialog::cellManuallyEdited(int row, int col) {
	
	if (row < 0 || col < 0 || row >= tableScheme->numRows() || col >= tableScheme->numCols()) {
		return;// Invalid selection, so we just return
	}
	
	assert(mainMixPtr != NULL);
	
	CellData *tempPointer = tableEntries[row][col]; // <-- save the old way the cell was, in case we want to revert it
	
	// Now let's parse the cell (the return value might be NULL, or a pointer to allocated CellData memory)
	try {
		tableEntries[row][col] = MouseIO::stringToCellData(tableScheme->text(row, col), mainMixPtr);
		
		if (tempPointer) {
			// And now we delete the old CellData, since we successfully allocated a new CellData
			delete tempPointer;
		}
		
	} catch (const string* e) {
		// Something happened while decoding the cell--maybe the user input a bad string.
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
		// So we pop up the box, and then return the cell to the way it was before they changed it
		tableEntries[row][col] = tempPointer; // ok... return it back to the way it was	
		// updateSchemeInfo should take care of setting the cell to the proper value
	}
	
	makeCurrentPaneDirty();
	updateSchemeInfo(row, col);
}



void SettingsDialog::reject(void) {
	// nothing... (occurs if user hits the escape key: default behavior is to just kill the window, without asking the user whether or not to save changes)
}

void SettingsDialog::closeEvent(QCloseEvent* closer) {
	if (!settingsDirty && !markersDirty) {	
		closer->accept();  // everything is clean, so no need to save changes
		return;
	}
	
	// otherwise...
	
	switch(QMessageBox::warning( this, "GenomeMixer",
								 "Should we really close this window now?"
								 "You have unapplied changes, and\n"
								 "if you close this dialog without applying\n"	
								 "them, they will not be saved.",
								 "Close without applying", "Cancel", "Apply changes",
								 2, // the "enter" button hits option 2 (apply changes)
								 1 // the escape button hits option 1 (cancel)
								 )) {
		case 0:   // the "forget changes" button
		closer->accept();	// <-- close the window without saving
		break;
		case 1:  // the "cancel" button
		closer->ignore();  // <-- cancels the "close" operation
		//      (not 100% guaranteed in X11--some window managers might still kill the window)
		break;
		case 2:  // the "apply changes" button
		applyChanges();   // <-- save changes first...
		closer->accept(); // <-- then close the window
		break;
	default:
		closer->ignore();
		break;
	}
	
}



void SettingsDialog::spinTotalChanged( int value ) {
	int row = tableScheme->currentRow();
	int col = tableScheme->currentColumn();
	if (row < 0 || col < 0) return;
	
	this->makeCellWhenThereIsNone(row, col);
	
	// Called when the "randomly assign sex for these progeny" spinner changes value
	
	schemeSpinRandomSex->setValue(value);
	tableEntries[row][col]->setRandomSex(value);
	
	makeCurrentPaneDirty();
	updateSchemeInfo();
}

void SettingsDialog::spinFemalesChanged( int value ) {	
	int row = tableScheme->currentRow();
	int col = tableScheme->currentColumn();
	if (row < 0 || col < 0) return;
	
	this->makeCellWhenThereIsNone(row, col);
	
	// Called when the "at least this many female mice" (progeny) spinner changes value
	schemeSpinFem->setValue(value);
	tableEntries[row][col]->setMinFemales(value);
	
	makeCurrentPaneDirty();
	updateSchemeInfo();
}



void SettingsDialog::spinMalesChanged( int value ) {
	int row = tableScheme->currentRow();
	int col = tableScheme->currentColumn();
	if (row < 0 || col < 0) return;
	
	this->makeCellWhenThereIsNone(row, col);
	
	// Called when the "at least this many male mice" (progeny) spinner changes value 
	schemeSpinMale->setValue(value);
	tableEntries[row][col]->setMinMales(value);
	
	makeCurrentPaneDirty();
	updateSchemeInfo();    
}



void SettingsDialog::schemeParentPulldownFemale( int item ) {
	this->schemeParentPulldown(item, kFEMALE);
}
void SettingsDialog::schemeParentPulldownMale( int item ) {
	this->schemeParentPulldown(item, kMALE);
}
void SettingsDialog::schemeParentPulldown( int item, bool sex) {
	// item is "which item was selected from this menu" and sex is "was it the female pulldown or the male pulldown"
	int row = tableScheme->currentRow();
	int col = tableScheme->currentColumn();
	if (row < 0 || col < 0) return;
	
	this->makeCellWhenThereIsNone(row, col);
	
	// Called when the pull-down for selecting the male parent is selected
	
	if (item == pulldown_purebred) {
		// If the user selected "Purebred..." from the pulldown:
		int thePurebredSelected;
		ChoosePurebred *cp = new ChoosePurebred();
		cp->setUpPurebreds(mainMixPtr, &thePurebredSelected); // <-- important!	
		// Now open the new dialog, and see what the user selects (it changes the value of &thePurebredSelected)
		if (cp->exec() == QDialog::Accepted
			&& thePurebredSelected >= 0 
			&& thePurebredSelected < mainMixPtr->numParentalStrains()) {
			
			if (tableEntries[row][col] == NULL) {
				tableEntries[row][col] = new CellData();
			}
			
			if (sex == kFEMALE) {
				tableEntries[row][col]->setMotherPurebred(thePurebredSelected);
			} else {
				tableEntries[row][col]->setFatherPurebred(thePurebredSelected);
			}   
			
		} else {
			// do nothing... either the user cancelled, or they didn't select a proper parental strain
			return;
		}
		
		delete cp;	// delete the "ChoosePurebred" dialog...
		
	} else if (item == pulldown_from_cell) {
		// Well, if they pull down the menu to "from cell...," then we do this:
		if (sex == kFEMALE) {
			tableEntries[row][col]->setMother(0, 0);
		} else {
			tableEntries[row][col]->setFather(0, 0);
		}
	} else {
		cerr << "The pulldown menu was pulled down to a value that we don't understand." << endl;
	}   
	
	makeCurrentPaneDirty();
	updateSchemeInfo();
}




void SettingsDialog::schemeFemaleParentManualEdit() {
	// Called when the user hits return after fiddling with the "female parent" string manually
	int row = tableScheme->currentRow();
	int col = tableScheme->currentColumn();
	if (row < 0 || col < 0) return;
	
	this->makeCellWhenThereIsNone(row, col);
	
	int cagePtr, genPtr, purebredIndexPtr;
	
	try {
		MouseIO::parentStrDecode(schemeParentFem->text(), mainMixPtr, &cagePtr, &genPtr, &purebredIndexPtr);
	} catch (const string* e) {
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
		return;
	}
	if (purebredIndexPtr == -1) {
		// Not a purebred...
		tableEntries[row][col]->setMother(cagePtr, genPtr);
	} else {
		// it is a purebred...
		try {
			tableEntries[row][col]->setMotherPurebred(purebredIndexPtr);
		} catch (const string* e) {
			QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
			return;
		}
	}
	
	makeCurrentPaneDirty(); // we changed the "enabled" status
	updateSchemeInfo();
	textDisplayInfo->setText("Your changes were entered into the table,\nalthough they have have not been applied yet.");
}




void SettingsDialog::schemeMaleParentManualEdit() {
	// Called when the user hits return after fiddling with the "male parent" string manually
	int row = tableScheme->currentRow();
	int col = tableScheme->currentColumn();
	if (row < 0 || col < 0) return;
	
	this->makeCellWhenThereIsNone(row, col);
	
	int cagePtr, genPtr, purebredIndexPtr;
	
	try {
		MouseIO::parentStrDecode(schemeParentMale->text(), mainMixPtr, &cagePtr, &genPtr, &purebredIndexPtr);
	} catch (const string* e) {
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
		return;
	}
	
	if (purebredIndexPtr == -1) {
		// Not a purebred...
		tableEntries[row][col]->setFather(cagePtr, genPtr);
	} else {
		// it is a purebred...
		try {
			tableEntries[row][col]->setFatherPurebred(purebredIndexPtr);
		} catch (const string* e) {
			QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
			return;
		}
	}
	
	makeCurrentPaneDirty(); // we changed the "enabled" status
	updateSchemeInfo();
	textDisplayInfo->setText("Your changes were entered into the table,\nalthough they have have not been applied yet.");
}




void SettingsDialog::schemeParentBeingManuallyEdited() {
	// Called when the user has changed either of the parent strings manually, but hasn't hit "return" to confirm
	textDisplayInfo->setText("Manually-entered changes to the parent won't\ntake effect until you hit return or enter.");
}



